# Copyright 2018 Carsten Blank
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
Qft
====

.. currentmodule:: dc_qiskit_algorithms.Qft

Here are a couple of functions that implement the quantum fourier transform:

.. autosummary::
   :nosignatures:

   get_theta
   qft
   qft_dg

See below for a description of the different functions to apply a quantum fourier transform

get_theta
##########

.. autofunction:: get_theta

qft
####

.. autofunction:: qft

qft_dg
#######

.. autofunction:: qft_dg

"""

import math
from typing import Tuple, List, Union

import qiskit.extensions.standard as standard
from qiskit import QuantumRegister, QuantumCircuit


def get_theta(k):
    # type: (int) -> float
    """
    As the conditional rotations are defined by a parameter k we have a
    convenience function for this: theta = (+/-) 2pi/2^(|k|)
    :param k: the integer for the angle
    :return: the angle
    """
    sign = -1 if k < 0 else 1
    lam = sign * 2 * math.pi * 2**(-abs(k))
    return lam


def qft(qc, q):
    # type: (QuantumCircuit, Union[List[Tuple[QuantumRegister, int]], QuantumRegister]) -> QuantumCircuit
    """
    Applies the Quantum Fourier Transform to q
    :param qc: the circuit to which the qft is applied
    :param q: the quantum register or list of quantum register/index tuples
    :return: the circuit with applied qft
    """
    q_list = []  # type: List[Tuple[QuantumRegister, int]]
    if isinstance(q, QuantumRegister):
        q_list = [q[i] for i in range(q.size)]
    else:
        q_list = q

    unused = q_list.copy()
    for qr in q_list:
        standard.h(qc, qr)
        k = 2
        unused.remove(qr)
        for qj in reversed(unused):
            standard.cu1(qc, get_theta(k), qj, qr)
            k = k + 1
    return qc


def qft_dg(qc, q):
    # type: (QuantumCircuit, Union[List[Tuple[QuantumRegister, int]], QuantumRegister]) -> QuantumCircuit
    """
        Applies the inverse Quantum Fourier Transform to q
        :param qc: the circuit to which the qft_dag is applied
        :param q: the quantum register or list of quantum register/index tuples
        :return: the circuit with applied qft_dag
        """
    qc2 = QuantumCircuit(*qc.qregs, *qc.cregs)
    qft(qc2, q)
    new_data = [op.inverse() for op in reversed(qc2.data)]
    qc2.data = new_data
    qc.extend(qc2)
    return qc



# Copyright 2018 Carsten Blank
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
UniformRotation
================

.. currentmodule:: dc_qiskit_algorithms.UniformRotation

This module implements the uniform rotation gate defined by Möttönen et. al.
(10.1103/PhysRevLett.93.130502).

There are many convenience functions also being monkey patched.

.. autosummary::
   :nosignatures:

   binary_codes
   gray_code
   matrix_M_entry
   compute_theta
   pairwise
   UniformRotationGate
   uni_rot
   uni_rot_dg
   unirz
   unirz_dg
   uniry
   uniry_dg
   MultiControlledXGate
   ccx
   ccx_dg

Here are the details:

binary_codes
#############

.. autofunction:: binary_codes

gray_code
##########

.. autofunction:: gray_code

matrix_M_entry
###############

.. autofunction:: matrix_M_entry

compute_theta
##############

.. autofunction:: compute_theta

pairwise
#########

.. autofunction:: pairwise

UniformRotationGate
####################

.. autoclass:: UniformRotationGate

uni_rot
########

.. autofunction:: uni_rot

uni_rot_dg
############

.. autofunction:: uni_rot_dg

unirz
######

.. autofunction:: unirz

unirz_dg
#########

.. autofunction:: unirz_dg

uniry
######

.. autofunction:: uniry

uniry_dg
#########

.. autofunction:: uniry_dg

MultiControlledXGate
#####################

.. autoclass:: MultiControlledXGate

ccx
#####

.. autofunction:: ccx

ccx_dg
#######

.. autofunction:: ccx_dg

"""
import logging
from itertools import tee
from typing import List, Tuple, Union, Callable, Iterable, Optional

import numpy as np
from qiskit import QuantumRegister, QuantumCircuit
from qiskit.circuit import CompositeGate, InstructionSet, Gate
from qiskit.circuit.register import Register
from qiskit.extensions.standard import ry, rz, cx
from scipy import sparse

log = logging.getLogger('UniformRotation')


def binary_codes(number_qubits):
    # type: (int) -> List[int]
    """
    Convenience function to get a list of numbers from 0 to 2**number_qubits - 1
    :param number_qubits: exponent
    :return: list of numbers
    """
    N = int(2**number_qubits)
    return list(range(N))


def gray_code(number):
    # type: (int) -> int
    """
    Cyclic Gray Code of number
    :param number: input number
    :return: Gray Code
    """
    return (number >> 1) ^ number


def matrix_M_entry(row, col):
    # type: (int, int) -> float
    """
    The matrix for the angle computation
    :param row: row number (one based!)
    :param col: column number (one based!)
    :return: matrix entry
    """
    b_and_g = row & gray_code(col)
    sum_of_ones = 0
    while b_and_g > 0:
        if b_and_g & 0b1:
            sum_of_ones += 1
        b_and_g = b_and_g >> 1
    return (-1)**sum_of_ones


def compute_theta(alpha):
    # type: (sparse.dok_matrix) -> sparse.dok_matrix
    """
    Compute the rotational angles from alpha
    :param alpha: the input uniform rotation angles
    :return: the single qubit rotation angles
    """
    k = np.log2(alpha.shape[0])
    factor = 2**(-k)

    theta = sparse.dok_matrix(alpha.shape, dtype=np.float64)  # type: sparse.dok_matrix
    for row in range(alpha.shape[0]):
        # Use transpose of M:
        entry = sum([matrix_M_entry(col, row) * a for (col, _), a in alpha.items()])
        entry *= factor
        if abs(entry) > 1e-6:
            theta[row, 0] = entry

    return theta


def pairwise(iterable):
    # type: (Iterable) -> Iterable[Tuple]
    """
    Calculates pairwise consecutive pairs of an iterable
    s -> (s0,s1), (s1,s2), (s2, s3), ...
    :param iterable: any iterable
    :return: an iterable of tuples
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class UniformRotationGate(CompositeGate):
    """Uniform rotation gate (Möttönen)."""

    def __init__(self, gate, alpha, control_qubits, tgt, circ=None):
        # type: (Callable[[CompositeGate, float, Tuple[QuantumRegister,int]], InstructionSet], sparse.dok_matrix, List[Tuple[QuantumRegister, int]], Tuple[QuantumRegister, int], Optional[QuantumCircuit]) -> None
        """
        Create new uniform rotation gate.
        :param gate: Either a ry or rz gate
        :param alpha: The conditional rotation angles
        :param control_qubits: control qubits
        :param tgt: the target qubit to apply the rotations
        :param circ: a circuit that this gate is being applied on
        """
        super().__init__("uni_rot_" + str(gate), [], control_qubits + [tgt], circ)

        theta = compute_theta(alpha)  # type: sparse.dok_matrix

        gray_code_rank = len(control_qubits)
        if gray_code_rank == 0:
            gate(self, theta[0, 0], tgt)
            return

        from sympy.combinatorics.graycode import GrayCode
        gc = GrayCode(gray_code_rank)  # type: GrayCode

        current_gray = gc.current
        for i in range(gc.selections):
            gate(self, theta[i, 0], tgt)
            next_gray = gc.next(i + 1).current

            control_index = int(np.log2(int(current_gray, 2) ^ int(next_gray, 2)))
            cx(self, control_qubits[control_index], tgt)

            current_gray = next_gray


def uni_rot(self, gate, alpha, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], Callable[[CompositeGate, float, Tuple[QuantumRegister,int]], InstructionSet], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply a generic uniform rotation with rotation gate.
    :param self: either a composite gate or a circuit
    :param gate: A single qubit rotation gate
    :param alpha: conditional rotation angles
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    if isinstance(control_qubits, QuantumRegister):
        instructions = InstructionSet()
        ctrs = [(control_qubits, j) for j in range(control_qubits.size)]
        if isinstance(tgt, QuantumRegister):
            for j in range(tgt.size):
                instructions.add(uni_rot(self, gate, alpha, ctrs, (tgt, j)))
        else:
            instructions.add(uni_rot(self, gate, alpha, ctrs, tgt))
        return instructions

    if not isinstance(alpha, sparse.dok_matrix):
        alpha = sparse.dok_matrix([alpha]).transpose()

    self._check_qubit(tgt)
    for qb in control_qubits:
        self._check_qubit(qb)
    return self._attach(UniformRotationGate(gate, alpha, control_qubits, tgt, self))


def uni_rot_dg(self, gate, alpha, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], Callable[[CompositeGate, float, Tuple[QuantumRegister,int]], InstructionSet], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply the dagger (inverse) of a generic uniform rotation with rotation gate.
    :param self: either a composite gate or a circuit
    :param gate: A single qubit rotation gate
    :param alpha: conditional rotation angles
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    return uni_rot(self, gate, alpha, control_qubits, tgt).inverse()


def unirz(self, alpha, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply a uniform rotation around z.
    :param self: either a composite gate or a circuit
    :param alpha: conditional rotation angles
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    def rz_möttönen2ibm(cg: CompositeGate, theta: float, qreg: Tuple[QuantumRegister, int]) -> InstructionSet:
        return rz(cg, -theta, qreg)

    return uni_rot(self, rz_möttönen2ibm, alpha, control_qubits, tgt)


def unirz_dg(self, alpha, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply dagger (inverse) of a uniform rotation around z.
    :param self: either a composite gate or a circuit
    :param alpha: conditional rotation angles
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    return unirz(self, alpha, control_qubits, tgt).inverse()


def uniry(self, alpha, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply a uniform rotation around y.
    :param self: either a composite gate or a circuit
    :param alpha: conditional rotation angles
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    def ry_möttönen2ibm(cg: CompositeGate, theta: float, qreg: Tuple[QuantumRegister, int]) -> InstructionSet:
        return ry(cg, -theta, qreg)

    return uni_rot(self, ry_möttönen2ibm, alpha, control_qubits, tgt)


def uniry_dg(self, alpha, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply the dagger (inverse) of a uniform rotation around y.
    :param self: either a composite gate or a circuit
    :param alpha: conditional rotation angles
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    return uniry(self, alpha, control_qubits, tgt).inverse()


def cnry(self, theta, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], float, Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply a multiple controlled y rotation on the target qubit.
    :param self: either a composite gate or a circuit
    :param theta: rotation angle
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    length = 2**len(control_qubits)
    alpha = sparse.dok_matrix((length, 1), dtype=np.float64)
    alpha[-1] = theta
    return uniry(self, alpha, control_qubits, tgt)


def cnry_dg(self, theta, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], float, Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply the dagger (inverse) of a multiple controlled y rotation on the target qubit.
    :param self: either a composite gate or a circuit
    :param theta: rotation angle
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    return cnry(self, theta, control_qubits, tgt).inverse()


class MultiControlledXGate(CompositeGate):
    """Multi-Controlled X-Gate (via Möttönen)."""

    def __init__(self, conditional_case, control_qubits, tgt, circ=None):
        # type: (int, Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[Register, int], QuantumRegister], Optional[QuantumCircuit]) -> None
        """
        Create a new multi-controlled X gate according to the conditional (binary) case
        :param conditional_case: binary representation of 0/1 control case
        :param control_qubits: control qubits
        :param tgt: target qubit
        :param circ: circuit this gate is applied to
        """
        super().__init__("ccx_uni_rot", [conditional_case], control_qubits + [tgt], circ)

        length = 2 ** len(control_qubits)
        alpha = sparse.dok_matrix((length, 1), dtype=np.float64)
        alpha[conditional_case] = np.pi
        from qiskit.extensions.standard import h
        h(self, tgt)
        uniry(self, alpha, control_qubits, tgt)
        h(self, tgt)

    def __repr__(self):
        """
        Representation of this object.
        :return: string representing this object
        """
        return "{}({}) {};".format(self.name, self.param, ["{}[{}]".format(q.name, i) for q, i in self.qargs])


def ccx(self, conditional_case, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], int, Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply a multi-controlled X gate depending on conditional binary representation
    :param self: either a composite gate or a circuit
    :param conditional_case: the controlled case (1 or 0) in binary
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    if isinstance(control_qubits, QuantumRegister):
        instructions = InstructionSet()
        ctrs = [(control_qubits, j) for j in range(control_qubits.size)]
        if isinstance(tgt, QuantumRegister):
            for j in range(tgt.size):
                instructions.add(ccx(self, conditional_case, control_qubits, (tgt, j)))
        else:
            instructions.add(ccx(self, conditional_case, control_qubits, tgt))
        return instructions

    self._check_qubit(tgt)
    for qb in control_qubits:
        self._check_qubit(qb)
    return self._attach(MultiControlledXGate(conditional_case, control_qubits, tgt, self))


def ccx_dg(self, conditional_case, control_qubits, tgt):
    # type: (Union[CompositeGate, QuantumCircuit], int, Union[List[Tuple[QuantumRegister, int]],QuantumRegister], Union[Tuple[QuantumRegister, int], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Apply the dagger (inverse) a multi-controlled X gate depending on conditional binary representation
    :param self: either a composite gate or a circuit
    :param conditional_case: the controlled case (1 or 0) in binary
    :param control_qubits: control qubits
    :param tgt: target
    :return: applied composite gate or circuit
    """
    return ccx(self, conditional_case, control_qubits, tgt).inverse()


QuantumCircuit.uniry = uniry
QuantumCircuit.uniry_dg = uniry_dg
CompositeGate.uniry = uniry
CompositeGate.uniry_dg = uniry_dg

QuantumCircuit.unirz = unirz
QuantumCircuit.unirz_dg = unirz_dg
CompositeGate.unirz = unirz
CompositeGate.unirz_dg = unirz_dg

QuantumCircuit.cnry = cnry
QuantumCircuit.cnry_dg = cnry_dg
CompositeGate.cnry = cnry
CompositeGate.cnry_dg = cnry_dg


# Copyright 2018 Carsten Blank
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
MöttönenStatePrep
==================

.. currentmodule:: dc_qiskit_algorithms.MöttönenStatePrep

This module implements the state preparation scheme defined by Möttönen et. al.
(http://dl.acm.org/citation.cfm?id=2011670.2011675)

.. autosummary::
   :nosignatures:

   get_alpha_z
   get_alpha_y
   state_prep_möttönen
   state_prep_möttönen_dg
   MöttönenStatePrep

See below the meaning and usage of the functions and the class

get_alpha_z
############

.. autofunction:: get_alpha_z

get_alpha_y
############

.. autofunction:: get_alpha_y

get_alpha_y
############

.. autofunction:: state_prep_möttönen

get_alpha_y
############

.. autofunction:: state_prep_möttönen_dg

MöttönenStatePrep
##################

.. autoclass:: MöttönenStatePrep

"""

import math
from typing import List, Tuple, Union, Optional

import numpy
from qiskit import QuantumRegister, QuantumCircuit
from qiskit.circuit import InstructionSet, CompositeGate, Gate
from scipy import sparse

from .UniformRotation import uniry, unirz


def get_alpha_z(omega, n, k):
    # type: (sparse.dok_matrix, int, int) -> sparse.dok_matrix
    """
    Computes the rotation angles alpha for the z-rotations
    :param omega: the input phase
    :param n: total number of qubits
    :param k: current qubit
    :return: a sparse vector
    """
    alpha_z_k = sparse.dok_matrix((2 ** (n - k), 1), dtype=numpy.float64)

    for (i, _), om in omega.items():
        i += 1
        j = int(numpy.ceil(i * 2 ** (-k)))
        s_condition = 2 ** (k - 1) * (2 * j - 1)
        s_i = 1.0 if i > s_condition else -1.0
        alpha_z_k[j - 1, 0] = alpha_z_k[j - 1, 0] + s_i * om / 2 ** (k - 1)

    return alpha_z_k


def get_alpha_y(a, n, k):
    # type: (sparse.dok_matrix, int, int) -> sparse.dok_matrix
    """
    Computes the rotation angles alpha for the y-rotations
    :param a: the input absolute values
    :param n: total number of qubits
    :param k: current qubit
    :return: a sparse vector
    """
    alpha = sparse.dok_matrix((2**(n - k), 1), dtype=numpy.float64)

    numerator = sparse.dok_matrix((2 ** (n - k), 1), dtype=numpy.float64)
    denominator = sparse.dok_matrix((2 ** (n - k), 1), dtype=numpy.float64)

    for (i, _), e in a.items():
        j = int(math.ceil((i + 1) / 2**k))
        l = (i + 1) - (2*j - 1)*2**(k-1)
        is_part_numerator = 1 <= l <= 2**(k-1)

        if is_part_numerator:
            numerator[j - 1, 0] += e*e
        denominator[j - 1, 0] += e*e

    for (j, _), e in numerator.items():
        numerator[j, 0] = numpy.sqrt(e)
    for (j, _), e in denominator.items():
        denominator[j, 0] = 1/numpy.sqrt(e)

    pre_alpha= numerator.multiply(denominator)  # type: sparse.csr_matrix
    for (j, _), e in pre_alpha.todok().items():
        alpha[j, 0] = 2*numpy.arcsin(e)

    return alpha


class MöttönenStatePrep(CompositeGate):
    """Uniform rotation Y gate (Möttönen)."""

    def __init__(self, vector, qubits, circ=None):
        # type: (MöttönenStatePrep, sparse.dok_matrix, List[Tuple[QuantumRegister, int]], Optional[QuantumCircuit]) -> None
        """
        Create the composite gate for the Möttönen state preparation scheme with an input vector, which registers/qubits
        to apply it to, and the circuit (if any)
        :param vector: the input complex sparse vector
        :param qubits: the qubits that will be applied the routine to create the state given by the vector
        :param circ: the circuit to which this composite gate is applied to
        """
        super().__init__("state_prep_möttönen", [], qubits, circ)
        a = sparse.dok_matrix(vector.get_shape())  # type: sparse.dok_matrix
        omega = sparse.dok_matrix(vector.get_shape())  # type: sparse.dok_matrix
        for (i, j), v in vector.items():
            a[i, j] = numpy.absolute(v)
            omega[i, j] = numpy.angle(v)
        self.apply_rot_z(omega, qubits)
        self.apply_rot_y(a, qubits)
        self.inverse()

    def apply_rot_y(self, a, qubits):
        # type: (MöttönenStatePrep, sparse.dok_matrix, List[Tuple[QuantumRegister, int]]) -> None
        """
        Applies the cascade of y-uniform rotations to the qubits
        :param a: the sparse absolute value vector
        :param qubits: qubits to which the scheme are applied
        :return: None
        """
        n = int(math.log2(a.get_shape()[0]))
        for k in range(1, n + 1):
            alpha_y_k = get_alpha_y(a, n, k)  # type: sparse.dok_matrix
            control = qubits[k:]
            target = qubits[k - 1]
            uniry(self, alpha_y_k, control, target)

    def apply_rot_z(self, omega, qubits):
        # type: (MöttönenStatePrep, sparse.dok_matrix, List[Tuple[QuantumRegister, int]]) -> None
        """
        Applies the cascade of z-uniform rotations to the qubits
        :param omega: the sparse phase vector
        :param qubits: qubits to which the scheme are applied
        :return: None
        """
        n = int(math.log2(omega.get_shape()[0]))
        for k in range(1, n + 1):
            alpha_z_k = get_alpha_z(omega, n, k)
            control = qubits[k:]
            target = qubits[k - 1]
            if len(alpha_z_k) != 0:
                unirz(self, alpha_z_k, control, target)


def state_prep_möttönen(self, a, qubits):
    # type: (Union[CompositeGate, QuantumCircuit], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
    Convenience function to encapsulate the composite gate of the state preparation
    :param self: Composite Gate or Quantum circuit to apply this to
    :param a: the input vector
    :param qubits: the qubits to be transformed
    :return: gate or instruction set
    """
    if isinstance(qubits, QuantumRegister):
        instructions = InstructionSet()
        qb = [(qubits, j) for j in range(qubits.size)]
        instructions.add(state_prep_möttönen(self, a, qb))
        return instructions

    for qb in qubits:
        self._check_qubit(qb)
    if isinstance(a, sparse.dok_matrix):
        return self._attach(MöttönenStatePrep(a, qubits, self))
    else:
        return self._attach(MöttönenStatePrep(sparse.dok_matrix([a]).transpose(), qubits, self))


def state_prep_möttönen_dg(self, a, qubits):
    # type: (Union[CompositeGate, QuantumCircuit], Union[List[float], sparse.dok_matrix], Union[List[Tuple[QuantumRegister, int]], QuantumRegister]) -> Union[Gate, InstructionSet]
    """
        Convenience function to encapsulate the composite gate of the dagger of the state preparation
        :param self: Composite Gate or Quantum circuit to apply this to
        :param a: the input vector
        :param qubits: the qubits to be transformed
        :return: gate or instruction set
        """
    return state_prep_möttönen_dg(self, a, qubits).inverse()


QuantumCircuit.state_prep_möttönen = state_prep_möttönen
QuantumCircuit.state_prep_möttönen_dg = state_prep_möttönen_dg
CompositeGate.state_prep_möttönen = state_prep_möttönen
CompositeGate.state_prep_möttönen_dg = state_prep_möttönen_dg



# Copyright 2018 Carsten Blank
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""
DraperAdder
===========

.. currentmodule:: dc_qiskit_algorithms.DraperAdder

This is the legendary Draper adder (arXiv:quant-ph/0008033).

.. autosummary::
   :nosignatures:

   draper_adder

More details:

draper_adder
############

.. autofunction:: draper_adder

"""
from typing import Optional, Tuple

import qiskit
from qiskit import QuantumCircuit
from qiskit.circuit.measure import measure
from qiskit.extensions import standard

from . import Qft as qft


def draper_adder(input_a, input_b, length = None):
    # type: (int, int, Optional[int]) -> Tuple[QuantumCircuit, float]
    """
    The Draper adder (arXiv:quant-ph/0008033), provide a and b and make sure to define a size of
    a register that can hold a or b
    :param input_a: integer a
    :param input_b: integer b
    :param length: size of qubit registers
    :return: tuple of the circuit and the length of the register
    """
    a_01s = "{0:b}".format(input_a)
    b_01s = "{0:b}".format(input_b)
    length = max(len(a_01s), len(b_01s), length if length is not None else 0)
    a_01s = a_01s.zfill(length)
    b_01s = b_01s.zfill(length)

    a = qiskit.QuantumRegister(len(a_01s), "a")
    b = qiskit.QuantumRegister(len(b_01s), "b")
    c_a = qiskit.ClassicalRegister(len(a_01s), "c_a")
    c_b = qiskit.ClassicalRegister(len(b_01s), "c_b")
    qc = qiskit.QuantumCircuit(a, b, c_a, c_b, name='draper adder')

    standard.barrier(qc, a, b)

    for i, c in enumerate(a_01s):
        if c == '1':
            standard.x(qc, a[i])

    for i, c in enumerate(b_01s):
        if c == '1':
            standard.x(qc, b[i])

    qft.qft(qc, a)

    for b_index in reversed(range(b.size)):
        theta_index = 1
        for a_index in reversed(range(b_index + 1)):
            standard.cu1(qc, qft.get_theta(theta_index), b[b_index], a[a_index])
            theta_index += 1

    qft.qft_dg(qc, a)

    standard.barrier(qc, a, b)

    measure(qc, a, c_a)
    measure(qc, b, c_b)

    return qc, 2**length
